import * as fs from 'fs'
import * as path from 'path'
import {Worker} from "node:worker_threads";

// 使用for ，而不是map、forEach，在大量数据时可以提高性能
const deepClone = target => {
	if (target && (typeof target == 'object')) {
		if (Array.isArray(target)) {
			var arr = [];
			for (var i = 0; i < target.length; i++) {
				arr.push(deepClone(target[i]))
			}
			return arr;
		}
		var obj = {};
		var keys = Object.keys(target);
		for (var i = 0; i < keys.length; i++) {
			obj[keys[i]] = deepClone(target[keys[i]])
		}
		return obj;
	}
	return target;
}

global.nv_remove_reference = deepClone;

const readJSON = file_path => {
	var data = fs.readFileSync(file_path);
	return JSON.parse(data.toString());
}

export default class jsonDB {
	#cols = [];
	#indexName = "";
	#db = [];
	#index = 1;
	#filePath = "";
	#references = new Map();

	get db() {
		return this.#db;
	}
	get columns() {
		return [...this.#cols];
	}
	get index() {
		return this.#indexName;
	}
	constructor(filePath,config) {
		this.#filePath = filePath;
		this.write = this._createThrottleFunc(this._write, config.threshold)
		this._readDB(filePath,config);
		return this;
	}
	_manullyInit({columns,index}) {
		if (!columns instanceof Array || !index instanceof String) {throw new Error("Args Error");}
		this.#cols = [...columns];
		this.#indexName = index;
	}
	_handleTempFiles() {
		var dirPath = path.parse(this.#filePath).dir;
		var basename = path.parse(this.#filePath).base;

		try {
			// getTempList
			var tempList = fs.readdirSync(dirPath)
			.map(r=>{
				var reg = new RegExp(`^${basename}-temp-([0-9]*).json$`.replace(/\./g,'\\.'));
				var result = r.match(reg);
				return result ? parseInt(result[1]) : null;
			})
			.filter(r=>r)
			// getDbModifyTime
			var mTime = fs.statSync(this.#filePath).mtimeMs
			// tempList找出比mTime时间新的数据
			var newest = tempList
			.filter(r=>r > mTime)
			.sort((a,b)=>a-b)[0]
			if (newest) {
				// 尝试读取json，如果没有报错，就说明是正常的文件
				var newest_is_ok = false;
				try {
					var read_result = readJSON(`${this.#filePath}-temp-${newest}.json`);
					newest_is_ok = true;
					read_result = null;
				} catch(e) {}
				if (newest_is_ok) {
					fs.renameSync(`${this.#filePath}-temp-${newest}.json`, this.#filePath);
					console.log(`${basename} 存在上次未成功写入的缓存，且数据正常。已替换`)
				} else {
					console.log(`${basename} 在上次保存时出现异常，已使用旧数据。`)
				}
			}
			//删除剩下全部的
			tempList.forEach(r=>{
				fs.unlink(`${this.#filePath}-temp-${r}.json`,(e)=>{})
			})
		} catch(e){}
	}
	_readDB(filePath,config) {
		this._handleTempFiles();
		//同步读取
		//文件不存在
		var basename = path.parse(filePath).base;
		var jsonFileExists = null;
		try {
			var stat = fs.statSync(filePath);
			jsonFileExists = stat.isFile();
		} catch(e) {
			jsonFileExists = false;
		}
		
		if (!jsonFileExists) {
			console.log(`数据文件不存在，新建 ${basename}`)
			this._manullyInit(config);
			this.write();
			return;
		}
		
		try {
			var data = readJSON(filePath);
			if (config) {
				this.#cols = data.cols;
				this.#indexName = data.indexName;
				this.#db = data.db;
				this.#index = data.index;
				// 注意这是config，其他是data
				(config.references || []).forEach(el=>{
					this.#references.set(el,true); // 随便设置一个值，只要后面has获取的时候能取到就行了
				})

				var addCols = [];
				var removeCols = [];
				//需要新增的列
				config.columns.forEach(el => {
					if (data.cols.indexOf(el) == -1) {
						addCols.push(el)
					}
				})
				//需要删除的列
				data.cols.forEach(el => {
					if (config.columns.indexOf(el) == -1) {
						removeCols.push(el)
					}
				})

				if (
					addCols.length !== 0
					|| removeCols.length !== 0
					|| data.indexName !== config.index
					) {
					console.log(`${basename} 数据不匹配，修复数据：`)

					if (addCols.length > 0) {
						addCols.forEach(el=>{
							this.addColumn(el)
						})
						console.log('新增列：',addCols);
					}

					if (removeCols.length > 0) {
						removeCols.forEach(el=>{
							this.removeColumn(el)
						})
						console.log('删除列：',removeCols);
					}

					if (data.indexName !== config.index) {
						this.#indexName = config.index;
						this.write();
						console.log('IndexName已修改',config.index);
					}

				}
			}
		} catch(e) {
			//json数据读取错误，一般是格式问题
			var nowStamp = Date.now();
			fs.renameSync(filePath, `${filePath}-error-${nowStamp}.json`)
			console.log(`数据文件错误，旧文件已重命名为：${basename}-error-${nowStamp}.json，新建数据文件`)
			this._manullyInit(config);
			this.write();
		}
	}
	_write() {
		// 由于 JSON.stringify 对CPU性能开销大，因此使用子线程处理
		// const worker = 
		new Worker(path.join(import.meta.dirname,"./class-jsonDB-write.js"), {workerData: {
			cols: this.#cols,
			indexName: this.#indexName,
			db: this.#db,
			index: this.#index,
			filePath: this.#filePath,
		}});

		/*var content = JSON.stringify({
			cols: this.#cols,
			indexName: this.#indexName,
			db: this.#db,
			index: this.#index,
		})

		var basename = path.parse(this.#filePath).base;
		var nowStamp = Date.now();
		// 尝试写入temp
		fs.writeFile(`${this.#filePath}-temp-${nowStamp}.json`, content, err=>{
			if (err) {
				console.error(`数据 ${basename} 缓存写入失败！`,err);
				return;
			}
			fs.rename(`${this.#filePath}-temp-${nowStamp}.json`,`${this.#filePath}`,err=>{
				if (err) {
					console.error(`数据 ${basename} 缓存重命名失败！`,err);
				}
			})
		})*/
	}
	_createThrottleFunc(func, threshold = 4500) {
		var last;
		var timer;
		return function () {
			var context = this;
			var now = +new Date,
			args = arguments;
			if (last && now < last + threshold) {
				clearTimeout(timer);
				timer = setTimeout(function () {
					last = now;
					func.apply(context, args);
				}, threshold);
			} else {
				last = now;
				func.apply(context, args);
			}
		};
	}
	insert(obj) {
		var newObj = deepClone(obj);
		var id = this.#index++;

		/*this.#db.push([
			id,
			...this.#cols.map(colName => newObj[colName])
		]);*/

		var row = [id];
		for (var i = 0; i < this.#cols.length; i++) {
			row.push( newObj[this.#cols[i]] )
		}
		this.#db.push(row)
		
		this.write();
		return id;
	}
	_getRow(arr, use_reference) {
		var result = {
			[this.#indexName]: arr[0]
		};
		if (use_reference) {
			// 使用引用类型。不deep_clone，高性能
			for (var i = 0; i < this.#cols.length; i++) {
				result[this.#cols[i]] = arr[i+1]
			}
		} else {
			// 默认状态，解除引用类型
			for (var i = 0; i < this.#cols.length; i++) {
				result[this.#cols[i]] = this.#references.has(this.#cols[i]) ? deepClone(arr[i+1]) : arr[i+1]
			}
		}
		return result;
	}
	addColumn(columnName) {
		if (!columnName instanceof String ) {throw new Error("Args Error");}
		if ( this.#cols.includes(columnName) ) {throw new Error(`Add Column Error: Column ${columnName} Exists`);}
		this.#cols.push(columnName);
		this.#db.forEach(row=>{
			row.length = this.#cols.length + 1;
		})
		this.write();
		return this;
	}
	removeColumn(columnName) {
		if (!columnName instanceof String ) {throw new Error("Args Error");}
		if ( !this.#cols.includes(columnName) ) {throw new Error(`Remove Column Error: Column ${columnName} Not Exists`);}
		var index = this.#cols.indexOf(columnName);
		this.#cols.splice(index,1);
		this.#db.forEach(row=>{
			row.splice(index+1,1)
		})
		this.write();
		return this;
	}
	find(filterFunc,columnNames) {
		var result = [];
		this._matchFilter(filterFunc,(logicRow,index)=>{
			if (columnNames) {
				var row = {};
				row[this.#indexName] = logicRow[this.#indexName];
				/*columnNames.forEach(el=>{
					row[el] = logicRow[el];
				})*/
				for (var i = 0; i < columnNames.length; i++) {
					row[columnNames[i]] = logicRow[columnNames[i]]
				}
				result.push(row)
			} else {
				result.push(logicRow)
			}
		})
		return result.reverse();
	}
	drop(filterFunc) {
		var dropedIDs = [];
		this._matchFilter(filterFunc,(logicRow,index)=>{
			dropedIDs.push( logicRow[this.#indexName] )
			this.#db.splice(index, 1);
		})
		this.write();
		return dropedIDs;
	}
	setColumn(filterFunc,obj) {
		var newObj = deepClone(obj);
		this._matchFilter(filterFunc,(logicRow,index)=>{
			for (var columnName in newObj) {
				var columnIndex = this.#cols.indexOf(columnName);
				columnIndex !== -1 && ( this.#db[index][columnIndex + 1] = newObj[columnName])
			}
		})
		this.write();
		return this;
	}
	_matchFilter(filterFunc,callback) {
		if (filterFunc instanceof Function) {
			var matches = filterFunc;
		}
		else if (filterFunc instanceof Object) {
			var matches = (row)=>{
				var results = [];
				for(var key in filterFunc) {
					var fromDB = row[key];
					var needle = filterFunc[key];

					if (needle instanceof Array) {
						results.push(needle.includes(fromDB))
					} else {
						results.push(fromDB == needle)
					}
				}
				return results.length > 0 ? results.reduce((a,b)=>a&&b) : false;
			}
		}

		for (var index = this.#db.length - 1; index >= 0; index--) {
			var logicRow = this._getRow( this.#db[index], filterFunc._nvdb_reference );
			if ( matches(logicRow) ) {
				callback(logicRow, index)
			}
		}
	}
}