/*

向WordPress钩子致敬：

1、do_action：预埋钩子，每次把【同一个参数】拿给钩子的回调函数
do_action(钩子名称, 回调参数)

2、apply_filters：预埋过滤器：每次把【上一次过滤后的结果作为参数】拿给钩子的回调函数
过滤结果 = apply_filters(钩子名称, 要过滤的回调参数)

3、add_action：添加钩子动作，在预埋的钩子处联动操作
add_action(钩子名称, 回调函数, 优先级)

4、add_filter：添加钩子过滤器，在预埋的过滤器处处理参数，需返回数据
add_filter(钩子名称, 回调函数, 优先级)

5、优先级：默认是10，可添加多个相同钩子名称的动作或过滤器。优先级越高的越先执行，相同优先级根据添加的先后顺序执行

注意：
1、Action 和 Filter 的动作名称【不要重复】
2、和WP的区别是: 回调参数只是1个。因此不需要 add 的时候在优先级后面跟参数数量。若需要多个参数，请用Object来实现，如：

预埋：
var arg1 = 'xxx', arg2 = 'xxx';
do_action( 'action_name', {arg1,arg2} );

使用钩子：
add_action( 'action_name', ( {arg1,arg2} ) => {
	console.log( arg1, arg2 );
} );

*/

// 直接定义一个对象存储钩子
var _WP_H00K = {};
// 结构类似于：
// {
// 	hook_name_1: [
// 		[function,priority],
// 		[function,priority],
// 		[function,priority]
// 	],
// 	hook_name_2: [
// 		[function,priority]
// 	]
// }

var check_deprecated_hooks = tag => {
	var deprecated_hooks = {
		_nv_init: "async:_nv_init",
	}
	if (deprecated_hooks[tag]) {
		throw new Error(`${tag}已过时，现使用：${deprecated_hooks[tag]}`);
	}
}

export default ()=>{
	// 下面所有的添加、移除、执行方法都暴露在window下供其他使用
	global.add_filter = function(tag, callback, priority) {
		check_deprecated_hooks(tag)
		if (!tag) {return false;}
		if (typeof callback !== 'function') {return false;}
		if (priority == undefined || typeof priority !== 'number') {priority = 10;}
		if (!_WP_H00K[tag]) {_WP_H00K[tag] = [];}
		var hook = _WP_H00K[tag];
		var insertAt = 0;
		for (var i = 0; i < hook.length; i++) {
			if (hook[i][1] <= priority) {
				insertAt = i + 1;
			}
			//完全相同的action，不会多次添加和执行
			if (hook[i][0] === callback) { return false; }
		}
		hook.splice(insertAt,0,[callback,priority]);
		return true;
	}

	global.add_action = add_filter;

	global.remove_filter = function(tag, callback) {
		if (!tag) {return false;}
		if (typeof callback !== 'function') {return false;}
		if (!_WP_H00K[tag]) {return false;}
		var hook = _WP_H00K[tag];
		for (var i = 0; i < hook.length; i++) {
			if(hook[i][0] === callback) {
				hook.splice(i,1);
				return true;
			}
		}
		return false;
	}

	global.remove_action = remove_filter;

	global.do_action = function(tag) {
		//每次都接受相同的arg
		var callbacks = _WP_H00K[tag];
		if (!callbacks) { return false; }
		var args = [];
		for (var i = 1; i < arguments.length; i++) {
			args.push(arguments[i]);
		}
		for (var i = 0; i < callbacks.length; i++) {
			callbacks[i][0].apply(void 0,args);
		}
	}

	global.apply_filters = function(tag,filter_me) {
		//每次的arg从上次返回得到
		var callbacks = _WP_H00K[tag];
		if (!callbacks) { return filter_me; }
		var args = [];
		for (var i = 2; i < arguments.length; i++) {
			args.push(arguments[i]);
		}
		for (var i = 0; i < callbacks.length; i++) {
			filter_me = callbacks[i][0].apply(void 0,[filter_me].concat(args));
		}
		return filter_me;
	}

	// 异步动作
	global.do_async_action = function(tag) {
		//每次都接受相同的arg
		var callbacks = _WP_H00K[tag];
		if (!callbacks) {
			return new Promise((reslve)=>{
				reslve();
			})
		}
		var args = [];
		for (var i = 1; i < arguments.length; i++) {
			args.push(arguments[i]);
		}
		async function fn() {
			for (var i = 0; i < callbacks.length; i++) {
				await new Promise((resolve)=>{
					callbacks[i][0].apply(void 0,[resolve].concat(args));
				})
			}
		}
		return fn();
	}

	// 异步过滤器
	global.apply_async_filters = function(tag,filter_me) {
		//每次的arg从上次返回得到
		var callbacks = _WP_H00K[tag];
		if (!callbacks) { return filter_me; }
		var args = [];
		for (var i = 2; i < arguments.length; i++) {
			args.push(arguments[i]);
		}
		async function fn() {
			for (var i = 0; i < callbacks.length; i++) {
				filter_me = await new Promise((resolve)=>{
					callbacks[i][0].apply(void 0,[resolve, filter_me].concat(args));
				})
			}
			return filter_me;
		}
		return fn();
	}
}

export var resetHooks = ()=>{
	_WP_H00K = {
		_nv_init: _WP_H00K._nv_init
	};
}

/*add_filter("async:sample_filter_hook",(next, tobe_filtered, arg1, arg2)=>{
    // 这将在第二步执行
    console.log(`第二步：${tobe_filtered}`)
    tobe_filtered = tobe_filtered + arg2;
    next(tobe_filtered)
}, 10)

add_filter("async:sample_filter_hook",(next, tobe_filtered, arg1, arg2)=>{
    // 这将在第一步执行，哪怕延迟了1秒，第二步也不会提前执行
    setTimeout(()=>{
        console.log(`第一步：${tobe_filtered}`)
        tobe_filtered = tobe_filtered + arg1;
        next(tobe_filtered)
    },1000)
}, 3)

add_filter("async:sample_filter_hook",(next, tobe_filtered, arg1, arg2)=>{
    // 这将在第三步执行
   setTimeout(()=>{
        console.log(`第三步：${tobe_filtered}`)
        tobe_filtered = tobe_filtered + 'END';
        next(tobe_filtered)
   },500)
}, 11)

apply_async_filters("async:sample_filter_hook", "被过滤", "参数1", "参数2")
.then(val=>{
    console.log("示例异步钩子全部执行完毕，结果：", val)
})*/


/*add_action("async:sample_hook",(next, arg1, arg2)=>{
    // 这将在第二步执行
    console.log(`第二步：${arg2}`)
    next()
}, 10)

add_action("async:sample_hook",(next, arg1, arg2)=>{
    // 这将在第一步执行，哪怕延迟了1秒，第二步也不会提前执行
    setTimeout(()=>{
        console.log(`第一步：${arg1}`)
        next()
    },1000)
}, 3)

add_action("async:sample_hook",(next, arg1, arg2)=>{
    // 这将在第三步执行
   setTimeout(()=>{
        console.log(`第三步：${arg1}-${arg2}`)
        next()
   },500)
}, 11)

do_async_action("async:sample_hook", "参数1", "参数2")
.then(()=>{
    console.log("示例异步钩子全部执行完毕")
})*/